from .base import Base
from .default import Default

from .zip import Zip
from .gzip import Gzip
from .lz4 import LZ4
from .zlib import Zlib

from .image import Image
from .jpeg import Jpeg
from .png import Png

def from_name(codec_name: str, compresslevel: float) -> Base:
    compresslevel = min(1, max(0, compresslevel))
    if codec_name == 'default':
        return Default()
    elif codec_name == 'gzip':
        return Gzip(compresslevel)
    elif codec_name == 'zlib':
        return Zlib(compresslevel)
    elif codec_name == 'lz4':
        return LZ4(compresslevel)
    elif codec_name == 'jpeg':
        return Jpeg()
    elif codec_name == 'png':
        return Png()
    else:
        raise Exception('Unknown Codec')


from typing import *
import os, sys, io, time, random, traceback, json


import numpy as np
import pickle as msgpack
from PIL import Image as PILImage

from .base import Base

class Image(Base):
    def __init__(self, format: str):
        self._format = format
        super().__init__()

    def encode(self, array: np.ndarray) -> bytes:
        info = {
            'shape': array.shape,
            'dtype': str(array.dtype)
        }

        assert array.dtype == 'uint8'
        assert len(array.shape) >= 3
        assert array.shape[-1] == 3

        if len(array.shape) == 3:
            bs = io.BytesIO()
            img = PILImage.fromarray(array)
            img.save(bs, format=self._format)
            info['image'] = bs.getvalue()
        else:
            images = []
            for i, index in enumerate(np.ndindex(array.shape[:-3])):
                bs = io.BytesIO()
                img = PILImage.fromarray(array[index])
                img.save(bs, format=self._format)
                images.append(bs.getvalue())
        info['images'] = images         
        return msgpack.dumps(info)
    
    def decode(self, content: bytes) -> np.ndarray:
        info = msgpack.loads(content)
        if 'image' in info:
            img = PILImage.open(io.BytesIO(bytearray(info['image'])))
            img.reshape(info['shape'])
            return img
        else:
            array = np.zeros(info['shape'], info['dtype'])
            images = info['images']
            for i, index in enumerate(np.ndindex(info['shape'][:-3])):
                img = PILImage.open(io.BytesIO(bytearray(images[i])))
                arr = np.asarray(img)
                array[index] = arr
            return array

    


import os

from hub import config
from hub.log import logger


class TokenManager(object):
    """ manages access tokens """

    @classmethod
    def is_authenticated(cls):
        return os.path.getsize(config.TOKEN_FILE_PATH) > 10

    @classmethod
    def set_token(cls, token):
        logger.debug(
            "Putting the key {} into {}.".format(token, config.TOKEN_FILE_PATH)
        )
        with open(config.TOKEN_FILE_PATH, "w") as f:
            f.write(token)

    @classmethod
    def get_token(cls):
        logger.debug("Getting token...")
        if not os.path.exists(config.TOKEN_FILE_PATH):
            return None
        with open(config.TOKEN_FILE_PATH, "r") as f:
            token = f.read()
        logger.debug("Got the key {} from {}.".format(token, config.TOKEN_FILE_PATH))
        return token

    @classmethod
    def get_auth_header(cls):
        logger.debug("Constructing auth header...")
        token = cls.get_token()
        if token:
            return "Bearer {}".format(token)
        return None

    @classmethod
    def purge_token(cls):
        logger.debug("Purging token...")
        if os.path.isfile(config.TOKEN_FILE_PATH):
            os.remove(config.TOKEN_FILE_PATH)



import zlib

from .zip import Zip

class Zlib(Zip):
    def __init__(self, compresslevel: float):
        compresslevel = int(round(6 + (compresslevel - 0.5) * 3)) if compresslevel >= 0.5 else int(round(compresslevel * 12))
        super().__init__(zlib, compresslevel)


import gzip

from .zip import Zip

class Gzip(Zip):
    def __init__(self, compresslevel: float):
        compresslevel = int(min(9, round(compresslevel * 18)))
        super().__init__(gzip, compresslevel)
    


import psutil
from dask.cache import Cache
from dask.distributed import Client
from hub.log import logger

_client = None


def get_client():
    global _client
    if _client is None:
        _client = init()
    return _client


def init(
    token: str = "",
    cache=2e9,
    cloud=False,
    n_workers=None,
    memory_limit=None,
    processes=False,
):
    """Initializes cluster either local or on the cloud
        
        Parameters
        ----------
        token: str
            token provided by snark
        cache: float
            Amount on local memory to cache locally, default 2e9 (2GB)
        cloud: bool
            Should be run locally or on the cloud
        n_workers: int
            number of concurrent workers, default 5
    """
    if cloud:
        raise NotImplementedError
    else:
        n_workers = n_workers if n_workers is not None else psutil.cpu_count()
        memory_limit = (
            memory_limit
            if memory_limit is not None
            else psutil.virtual_memory().available
        )

        client = Client(
            n_workers=n_workers, processes=processes, memory_limit=memory_limit
        )

    cache = Cache(cache)
    cache.register()
    global _client
    _client = client
    return client



from .collections import dataset
from .collections import tensor
from .collections.dataset import load
from .collections.client_manager import init
import hub.config


def local_mode():
    hub.config.HUB_REST_ENDPOINT = hub.config.HUB_LOCAL_REST_ENDPOINT
